<?php

namespace Ulmus;

use Generator;

class EntityCollection extends \ArrayObject {
    
    public ? string $entityClass = null;
    
    public function filters(Callable $callback) : Generator
    {
        $idx = 0;
        
        foreach($this as $key => $item) {
            if ( $callback($item, $key, $idx) ) {
                $idx++;
                
                yield $key => $item;
            }
        }
    }
        
    public function filtersCollection(Callable $callback) : self
    {
        $collection = new static();
        
        foreach($this->filters($callback) as $item) {
            $collection->append($item);
        }
        
        return $collection;
    }
        
    public function iterate(Callable $callback) : array
    {
        $results = [];
        
        foreach($this as $item) {
            $results[] = $callback($item);
        }
        
        return $results;
    }
    
    public function removeOne($value, string $field, bool $strict = true) : ? object
    {
        foreach($this->search($value, $field, $strict) as $key => $item) {
            
            $this->offsetUnset($key);
            
            return $item;
        }
        
        return null;
    }
    
    public function remove($value, string $field, bool $strict = true) : array
    {
        $removed = [];
        
        foreach($this->search($value, $field, $strict) as $key => $item) {
            $this->offsetUnset($key);
            
            $removed[] = $item;
        }
        
        return $removed;
    }
    
    public function searchOne($value, string $field, bool $strict = true) : ? object 
    {
        # Returning first value only
        foreach($this->search($value, $field, $strict) as $item) {
            return $item;
        }
        
        return null;
    }
    
    public function search($value, string $field, bool $strict = true) : Generator 
    {
        foreach($this->filters(fn($v) => $strict ? $v->$field === $value : $v->$field == $value) as $key => $item) {
            yield $key => $item;
        }
    }
    
    public function column($field) : array
    {
        $list = [];
        
        foreach($this as $item) {
            $list[] = $item->$field;
        }
        
        return $list;
    }
    
    public function buildArray(string $keyColumn, /* string|callable|null */ $value = null) : array
    {
        $list = [];
        
        foreach($this as $item) {
            switch (true) {
                case is_null($value):
                    $list[] = $item->$keyColumn;
                break;
                
                case is_callable($value):
                    $list[$item->$keyColumn] = $value($item);
                break;
                
                case is_object($value):
                case is_string($value):
                    $value = (string) $value;
                    
                    $list[$item->$keyColumn] = $item->$value;
                break;
            }
        }
        
        return $list;
    }
    
    public function implode(string $glue, ? Callable $callback = null) : string 
    {
        $values = [];
        
        foreach($this as $item) {
            $values[] = $callback ? $callback($item) : (string) $item;
        }
        
        return implode($glue, $values);
    }
    
    public function toArray(bool $includeRelations = false) : array {
        $list = [];
        
        foreach($this as $entity) {
            $list[] = $entity->toArray($includeRelations);
        }
        
        return $list;
    }
        
    public function fromArray(array $datasets, ? string $entityClass = null) : self 
    {
        if ( ! ($this->entityClass || $entityClass) ) {
            throw new \Exception("An entity class name must be provided to be instanciated and populated before insertion into this collection.");
        }
        
        $className = $entityClass ?: $this->entityClass;
        
        foreach($datasets as $dataset) {
            $this->append( (new $className() )->fromArray($dataset) );
        }
        
        return $this;
    }
    
    public function mergeWith( /*array|EntityCollection*/ $datasets ) : self
    {
        if ( is_object($datasets) ) {
            $datasets = $datasets->getArrayCopy();
        }
        
        $this->exchangeArray( array_merge( $this->getArrayCopy(), $datasets ) );
        
        return $this;
    }
}
